#include <iostream>
#include <cmath>              // mathematical C library
#include <complex> 
#include "f2c.h"

#include <fstream>

#include <algorithm>             // defining the max() operator
#include <vector>           // STL vector templates
using namespace std;
#include "finrecipes.h"



#ifndef PI 
const double PI=3.141592653589793238462643;
#endif 

using namespace std;

double n(const double& z) {  // normal distribution function    
    return (1.0/sqrt(2.0*PI))*exp(-0.5*z*z);
};

double N(const double& z) {
    if (z >  6.0) { return 1.0; }; // this guards against overflow 
    if (z < -6.0) { return 0.0; };

    double b1 =  0.31938153; 
    double b2 = -0.356563782; 
    double b3 =  1.781477937;
    double b4 = -1.821255978;
    double b5 =  1.330274429; 
    double p  =  0.2316419; 
    double c2 =  0.3989423; 

    double a=fabs(z); 
    double t = 1.0/(1.0+a*p); 
    double b = c2*exp((-z)*(z/2.0)); 
    double n = ((((b5*t+b4)*t+b3)*t+b2)*t+b1)*t; 
    n = 1.0-b*n; 
    if ( z < 0.0 ) n = 1.0 - n; 
    return n; 
};

inline double f(const double& x, const double& y, 
		const double& aprime, const double& bprime,
		const double& rho) {
    double r = aprime*(2*x-aprime) + bprime*(2*y-bprime) + 2*rho*(x-aprime)*(y-bprime);
    return exp(r);
};

inline double sgn(const double& x) {  // sign function
    if (x>=0.0)  return 1.0;
    return -1.0;
};

double N(const double& a, const double& b, const double& rho) {
    if ( (a<=0.0) && (b<=0.0) && (rho<=0.0) ) { 
	double aprime = a/sqrt(2.0*(1.0-rho*rho));
	double bprime = b/sqrt(2.0*(1.0-rho*rho));
	double A[4]={0.3253030, 0.4211071, 0.1334425, 0.006374323}; 
	double B[4]={0.1337764, 0.6243247, 1.3425378, 2.2626645  };
	double sum = 0;
	for (int i=0;i<4;i++) {
	    for (int j=0; j<4; j++) {
		sum += A[i]*A[j]* f(B[i],B[j],aprime,bprime,rho);
	    };
	};
	sum = sum * ( sqrt(1.0-rho*rho)/PI);
	return sum;
    }
    else  if ( a * b * rho <= 0.0 ) {
	if ( ( a<=0.0 ) && ( b>=0.0 ) && (rho>=0.0) ) {
	    return N(a) - N(a, -b, -rho); 
	}
	else if ( (a>=0.0) && (b<=0.0) && (rho>=0.0) ) {
	    return N(b) - N(-a, b, -rho); 
	}
	else if ( (a>=0.0) && (b>=0.0) && (rho<=0.0) ) {
	    return N(a) + N(b) - 1.0 + N(-a, -b, rho); 
	};
    }
    else  if ( a * b * rho >= 0.0 ) {
	double denum = sqrt(a*a - 2*rho*a*b + b*b);
	double rho1 = ((rho * a - b) * sgn(a))/denum;
	double rho2 = ((rho * b - a) * sgn(b))/denum;
	double delta=(1.0-sgn(a)*sgn(b))/4.0;
	return N(a,0.0,rho1) + N(b,0.0,rho2) - delta;
    };
    return -99.9; // should never get here, alternatively throw exception
};


double option_price_call_black_scholes(const double& S,       // spot (underlying) price
				       const double& K,       // strike (exercise) price,
				       const double& r,       // interest rate
				       const double& sigma,   // volatility 
				       const double& time) {  // time to maturity 
    double time_sqrt = sqrt(time);
    double d1 = (log(S/K)+r*time)/(sigma*time_sqrt)+0.5*sigma*time_sqrt; 
    double d2 = d1-(sigma*time_sqrt);
    return S*N(d1) - K*exp(-r*time)*N(d2);
};

double option_price_delta_call_black_scholes(const double& S,     // spot price
					     const double& K,     // Strike (exercise) price,
					     const double& r,     // interest rate
					     const double& sigma, // volatility
					     const double& time){  // time to maturity
    double time_sqrt = sqrt(time);
    double d1 = (log(S/K)+r*time)/(sigma*time_sqrt) + 0.5*sigma*time_sqrt; 
    double delta = N(d1);
    return delta;
};

void option_price_partials_call_black_scholes( const double& S,     // spot price
					       const double& K,     // Strike (exercise) price,
					       const double& r,     // interest rate
					       const double& sigma, // volatility
					       const double& time,  // time to maturity
					       double& Delta, //  partial wrt S
					       double& Gamma, //  second prt wrt S
					       double& Theta, // partial wrt time
					       double& Vega,  //  partial wrt sigma
					       double& Rho){   // partial wrt r
  double time_sqrt = sqrt(time);
  double d1 = (log(S/K)+r*time)/(sigma*time_sqrt) + 0.5*sigma*time_sqrt; 
  double d2 = d1-(sigma*time_sqrt);
  Delta = N(d1);
  Gamma = n(d1)/(S*sigma*time_sqrt);
  Theta =- (S*sigma*n(d1))/(2*time_sqrt) - r*K*exp( -r*time)*N(d2);
  Vega  = S * time_sqrt*n(d1);
  Rho   = K*time*exp(-r*time)*N(d2);
};

double option_price_put_black_scholes(const double& S,      // spot price
				      const double& K,      // Strike (exercise) price,
				      const double& r,      // interest rate
				      const double& sigma,  // volatility
				      const double& time){
    double time_sqrt = sqrt(time);
    double d1 = (log(S/K)+r*time)/(sigma*time_sqrt) + 0.5*sigma*time_sqrt;
    double d2 = d1-(sigma*time_sqrt);
    return K*exp(-r*time)*N(-d2) - S*N(-d1);
};

double option_price_delta_put_black_scholes(const double& S, // spot price
					    const double& K, // Strike (exercise) price,
					    const double& r,  // interest rate
					    const double& sigma,
					    const double& time) {
    double time_sqrt = sqrt(time);
    double d1 = (log(S/K)+r*time)/(sigma*time_sqrt) + 0.5*sigma*time_sqrt; 
    double delta = -N(-d1);
    return delta;
}
void option_price_partials_put_black_scholes(const double& S, // spot price
					     const double& K, // Strike (exercise) price,
					     const double& r,  // interest rate
					     const double& sigma,
					     const double& time,
					     double& Delta, //  partial wrt S
					     double& Gamma, //  second prt wrt S
					     double& Theta,  // partial wrt time
					     double& Vega,  //  partial wrt sigma
					     double& Rho){    // partial wrt r
    double time_sqrt = sqrt(time);
    double d1 = (log(S/K)+r*time)/(sigma*time_sqrt) + 0.5*sigma*time_sqrt; 
    double d2 = d1-(sigma*time_sqrt);
    Delta = -N(-d1);
    Gamma = n(d1)/(S*sigma*time_sqrt);
    Theta = -(S*sigma*n(d1)) / (2*time_sqrt)+ r*K * exp(-r*time) * N(-d2); 
    Vega  = S * time_sqrt * n(d1);
    Rho   = -K*time*exp(-r*time) * N(-d2);
};

double option_price_implied_volatility_call_black_scholes_bisections(const double& S,
								     const double& K,
								     const double& r,
								     const double& time,
								     const double& option_price){
    if (option_price<0.99*(S-K*exp(-time*r))) {  // check for arbitrage violations. 
	return 0.0;                             // Option price is too low if this happens
    };
  
    // simple binomial search for the implied volatility.
    // relies on the value of the option increasing in volatility
    const double ACCURACY = 1.0e-5; // make this smaller for higher accuracy
    const int MAX_ITERATIONS = 100;
    const double HIGH_VALUE = 1e10;
    const double ERROR = -1e40;
  
    // want to bracket sigma. first find a maximum sigma by finding a sigma
    // with a estimated price higher than the actual price.
    double sigma_low=1e-5;
    double sigma_high=0.3;
    double price = option_price_call_black_scholes(S,K,r,sigma_high,time);
    while (price < option_price) {  
	sigma_high = 2.0 * sigma_high; // keep doubling.
	price = option_price_call_black_scholes(S,K,r,sigma_high,time);
	if (sigma_high>HIGH_VALUE) return ERROR; // panic, something wrong.
    };
    for (int i=0;i<MAX_ITERATIONS;i++){
	double sigma = (sigma_low+sigma_high)*0.5;
	price = option_price_call_black_scholes(S,K,r,sigma,time);
	double test =  (price-option_price);
	if (fabs(test)<ACCURACY) { return sigma; };
	if (test < 0.0) { sigma_low = sigma; }
	else { sigma_high = sigma; }
    };
    return ERROR;
};

double option_price_implied_volatility_call_black_scholes_newton(const double& S,
								 const double& K, 
								 const double& r,
								 const double& time,
								 const double& option_price) {
    if (option_price<0.99*(S-K*exp(-time*r))) {  // check for arbitrage violations. Option price is too low if this happens
	return 0.0;
    };

    const int MAX_ITERATIONS = 100;
    const double ACCURACY    = 1.0e-5; 
    double t_sqrt = sqrt(time);

    double sigma = (option_price/S)/(0.398*t_sqrt);    // find initial value
    for (int i=0;i<MAX_ITERATIONS;i++){
	double price = option_price_call_black_scholes(S,K,r,sigma,time);
	double diff = option_price -price;
	if (fabs(diff)<ACCURACY) return sigma;
	double d1 = (log(S/K)+r*time)/(sigma*t_sqrt) + 0.5*sigma*t_sqrt; 
	double vega = S * t_sqrt * n(d1);
	sigma = sigma + diff/vega;
    };
    return -99e10;  // something screwy happened, should throw exception
};

double option_price_european_lookback_call(const double& S,
					   const double& Smin,
					   const double& r,
					   const double& q,
					   const double& sigma,
					   const double& time){
    if (r==q) return 0;
    double sigma_sqr=sigma*sigma;
    double time_sqrt = sqrt(time);
    double a1 = (log(S/Smin) + (r-q+sigma_sqr/2.0)*time)/(sigma*time_sqrt);
    double a2 = a1-sigma*time_sqrt;
    double a3 = (log(S/Smin) + (-r+q+sigma_sqr/2.0)*time)/(sigma*time_sqrt);
    double Y1 = 2.0 * (r-q-sigma_sqr/2.0)*log(S/Smin)/sigma_sqr;
    return S * exp(-q*time)*N(a1)- S*exp(-q*time)*(sigma_sqr/(2.0*(r-q)))*N(-a1)
	- Smin * exp(-r*time)*(N(a2)-(sigma_sqr/(2*(r-q)))*exp(Y1)*N(-a3));
};

double 
option_price_asian_geometric_average_price_call(const double& S,
						const double& K,
						const double& r,
						const double& q,
						const double& sigma,
						const double& time){
    double sigma_sqr = pow(sigma,2);
    double adj_div_yield=0.5*(r+q+sigma_sqr/6.0);
    double adj_sigma=sigma/sqrt(3.0);
    double adj_sigma_sqr = pow(adj_sigma,2);
    double time_sqrt = sqrt(time);
    double d1 = (log(S/K) + (r-adj_div_yield + 0.5*adj_sigma_sqr)*time)/(adj_sigma*time_sqrt);
    double d2 = d1-(adj_sigma*time_sqrt);
    double call_price = S * exp(-adj_div_yield*time)* N(d1) - K * exp(-r*time) * N(d2);
    return call_price;
};


struct heston_parms {double K; double x; double r;  double v; double tau;  double kappa; double theta; 
                     double rho; double sigma; double lambda;  int j;};




extern "C"  int iqlvms_(integer *nm, integer *n, doublereal *d__, 
	doublereal *e, doublereal *z__, integer *iev, integer *ier);

inline int eig_tridi(int nev,
					 int size, 
					 double *maindiag, 
					 double *subdiag,
					 double *ev)
{
  integer n = nev;
  integer ierr=0;
  integer sz = size;
  iqlvms_(&n, &sz, maindiag, subdiag, ev, &n, &ierr);
  return ierr;
};




extern "C"{


double heston_integrand_j(double phi, void *p){ 
    struct heston_parms* parms = (struct heston_parms*)p;
    double K = (parms->K); double x = (parms->x); 
    double v = (parms->v); double r = (parms->r);
    double kappa = (parms->kappa);
    double theta = (parms->theta);
    double rho = (parms->rho);
    double sigma = (parms->sigma);
    double lambda = (parms->lambda);
    double tau = (parms->tau);
    double j = (parms->j);
    double sigma_sqr = pow(sigma,2);
    double uj;    double bj;
    if (j==1){	uj=0.5;  bj=kappa+lambda-rho*sigma; }
    else {      uj=-0.5; bj=kappa+lambda;   };
    std::complex <double> i(0,1);
    double a = kappa*theta;
    std::complex<double> d  = sqrt( pow(rho*sigma*phi*i-bj,2) - sigma_sqr*(2*uj*phi*i-pow(phi,2)) );
    std::complex<double> g  = (bj - rho*sigma*phi*i+d)/(bj-rho*sigma*phi*i-d);
    std::complex<double> C  = r*phi*i*tau+(a/sigma_sqr)*((bj-rho*sigma*phi*i+d)*tau-2.0*log((1.0-g*exp(d*tau))/(1.0-g)));
    std::complex<double> D  = (bj-rho*sigma*phi*i+d)/sigma_sqr * ( (1.0-exp(d*tau))/(1.0-g*exp(d*tau)) );
    std::complex<double> f1 = exp(C+D*v+i*phi*x);
    std::complex<double> F  = exp(-phi*i*log(K))*f1/(i*phi);
	return std::real(F); 
};};


////USING FORTRAN
 static double x[150] = {
0.542626,0.370395,0.230993,0.124405,0.050618,0.009607,
0.747706,0.985657,1.256506,1.560281,
1.897016,2.266749,2.669519,3.105372,3.574356,4.076522,4.611926,5.180628,5.782691,6.418183,
7.087175,7.789743,8.525966,9.295927,10.099714,10.937420,11.809140,12.714974,13.655029,14.629413,
15.638241,16.681630,17.759705,18.872593,20.020428,21.203347,22.421494,23.675016,24.964068,26.288808,
27.649400,29.046015,30.478829,31.948023,33.453785,34.996309,36.575795,38.192451,39.846489,41.538130,
43.267601,45.035137,46.840980,48.685380,50.568593,52.490886,54.452532,56.453814,58.495024,60.576461,
62.698436,64.861268,67.065287,69.310833,71.598257,73.927922,76.300199,78.715476,81.174150,83.676631,
86.223344,88.814727,91.451232,94.133325,96.861491,99.636228,102.458051,105.327495,108.245112,
111.211471,114.227165,117.292806,120.409026,123.576484,126.795859,130.067858,133.393212,136.772683,
140.207058,143.697158,147.243835,150.847975,154.510500,158.232369,162.014582,165.858181,169.764254,
173.733934,177.768406,181.868909,186.036738,190.273249,194.579863,198.958070,203.409435,207.935600,
212.538294,217.219337,221.980648,226.824250,231.752283,236.767009,241.870826,247.066279,252.356069,
257.743073,263.230360,268.821204,274.519111,280.327841,286.251434,292.294241,298.460962,304.756684,
311.186934,317.757729,324.475644,331.347894,338.382420,345.588004,352.974402,360.552509,368.334558,
376.334374,384.567687,393.052539,401.809802,410.863871,420.243586,429.983502,440.125674,450.722256,
461.839398,473.563384,486.010824,499.346798,513.820397,529.844189,548.212947,570.989411};

static double w[150] = {
0.188652,0.155814,0.122992,0.090185,0.057392,0.024654,0.221512,0.254395,0.287307,0.320249,0.353228,
0.386245,0.419304,0.452410,0.485566,0.518775,0.552043,0.585372,0.618766,0.652230,0.685767,0.719382,
0.753078,0.786859,0.820731,0.854696,0.888760,0.922927,0.957201,0.991586,1.026088,1.060711,1.095460,
1.130339,1.165353,1.200509,1.235809,1.271261,1.306869,1.342638,1.378575,1.414685,1.450973,1.487446,
1.524110,1.560971,1.598036,1.635311,1.672802,1.710518,1.748464,1.786649,1.825079,1.863763,1.902708,
1.941923,1.981416,2.021196,2.061272,2.101654,2.142349,2.183370,2.224725,2.266426,2.308483,2.350908,
2.393712,2.436908,2.480508,2.524525,2.568974,2.613867,2.659220,2.705048,2.751367,2.798193,2.845543,
2.893437,2.941891,2.990927,3.040563,3.090823,3.141728,3.193301,3.245567,3.298552,3.352283,3.406789,
3.462099,3.518244,3.575258,3.633176,3.692034,3.751871,3.812729,3.874651,3.937683,4.001874,4.067277,
4.133945,4.201939,4.271320,4.342156,4.414519,4.488484,4.564134,4.641557,4.720849,4.802110,4.885451,
4.970991,5.058860,5.149198,5.242157,5.337903,5.436618,5.538501,5.643771,5.752670,5.865463,5.982447,
6.103950,6.230339,6.362026,6.499475,6.643209,6.793823,6.951998,7.118515,7.294276,7.480333,7.677916,
7.888481,8.113766,8.355859,8.617311,8.901266,9.211659,9.553501,9.933297,10.359677,10.844413,11.404090,
12.063035,12.858816,13.853485,15.159514,17.010660,20.015419,26.688256};

inline double heston_Pj(double S, double K, double r, double v, double tau, double sigma, 
                 double kappa, double lambda, double rho, double theta, int j){

	//FILE   *out_abs= fopen("abs.res","w"), *out_wexp= fopen("wexp.res","w");
    double y=log(S);
    struct heston_parms parms = { K, y, r, v, tau, kappa, theta, rho, sigma, lambda, j};
    size_t n=10000;

	int i;

	const int       glsteps = 150; /* <= 150 for double precision */


	///* Gauss-Laguerre */
	//double * x;  /* coordinates */
    //double * w;  /* weights */
	//  double *sub;      /* subdiagonal */
	//  w = (double*) calloc(sizeof(double), glsteps);
	//  x = (double*) calloc(sizeof(double), glsteps);
	//  sub = (double*) calloc(sizeof(double), glsteps);
	//  for (i=1; i<glsteps; ++i) w[i]=0.0;
	//  w[0]=1.0; 
	//  for (i=0; i<glsteps; ++i) x[i]=2*i+1;
	//  for (i=0; i<glsteps; ++i) sub[i]=i;
	//  i = eig_tridi( 1, glsteps, x, sub, w );
	//  for (i=0; i<glsteps; ++i){
	//	  w[i] *= w[i] * exp(x[i]);
	//	  //fprintf(out_abs, "%lf,", x[i]);
	//	  //fprintf(out_wexp, "%lf,", w[i]);
	//  }
	//  free(sub);
 //   //fclose(out_abs);
 //   //fclose(out_wexp); 
   
	/*integral*/
	double sum = 0, phi;
	for (i=glsteps-1;i>=0;--i)
	  {
		phi = x[i];
		sum += w[i] * heston_integrand_j(phi, &parms);
	  }

    return 0.5 + sum/PI;
};


/////NR prendre glsteps = 64 ou moins
//extern "C"{ void gaulag(float x[], float w[], int n, float alf);
//}
//
//inline double heston_Pj(double S, double K, double r, double v, double tau, double sigma, 
//                 double kappa, double lambda, double rho, double theta, int j){
//
//    double y=log(S);
//    struct heston_parms parms = { K, y, r, v, tau, kappa, theta, rho, sigma, lambda, j};
//	FILE *out_abs= fopen("abs.res","w"), *out_wexp= fopen("wexp.res","w");
//
//    const int glsteps = 64; /* <= 150 for double precision */
//
//	/* Gauss-Laguerre */
//	int i;
//	float alf=0.0,checkw,checkx,xx,
//		*x=(float*) calloc(sizeof(double), glsteps+1), /* coordinates */
//		*w=(float*) calloc(sizeof		(double), glsteps+1); /* weights */
//		gaulag(x,w,glsteps,alf);  
//	//for (i=1; i<glsteps +1; ++i){
//	//	  w[i] *= exp(x[i]);
//	//	  fprintf(out_abs, "%lf,", x[i]);
//	//	  fprintf(out_wexp, "%lf,", w[i]);
//	//  }
//    fclose(out_abs);
//    fclose(out_wexp); 
//
//	double sum = 0, phi;
//	for (i=glsteps-1;i>=0;--i)
//	  {
//		w[i+1] *=exp(x[i+1]);
//		phi = x[i+1];
//		sum += w[i+1] * heston_integrand_j(phi, &parms);
//	  }
//	free(x); 
//	free(w);
//    return 0.5 + sum/PI;
//};

	double heston_call_option_price(const double& S, const double& K, const double& r, const double& v, 
				const double& tau, const double& rho,  const double& kappa, 
				const double& lambda,  const double& theta, const double& sigma){
    double P1 = heston_Pj(S,K,r,v,tau,sigma,kappa,lambda,rho,theta,1);
    double P2 = heston_Pj(S,K,r,v,tau,sigma,kappa,lambda,rho,theta,2);
    double C=S*P1-K*exp(-r*tau)*P2;
    return C;
};



double option_price_call_merton_jump_diffusion( const double& S, 
						const double& X,
						const double& r,
						const double& sigma, 
						const double& time_to_maturity,
						const double& lambda,
						const double& kappa,
						const double& delta) {
    const int MAXN=50;
    double tau=time_to_maturity;
    double sigma_sqr = sigma*sigma; 
    double delta_sqr = delta*delta;
    double lambdaprime = lambda * (1+kappa);
    double gamma = log(1+kappa);
    double c = exp(-lambdaprime*tau)*option_price_call_black_scholes(S,X,r-lambda*kappa,sigma,tau);
    double log_n = 0;
    for (int n=1;n<=MAXN; ++n) {
	log_n += log(double(n));
	double sigma_n = sqrt( sigma_sqr+n*delta_sqr/tau );
	double r_n = r-lambda*kappa+n*gamma/tau;
	c += exp(-lambdaprime*tau+n*log(lambdaprime*tau)-log_n)*
	    option_price_call_black_scholes(S,X,r_n,sigma_n,tau);
    };
    return c;
};

double term_structure_discount_factor_vasicek(const double& time,
					      const double& r,
					      const double& a,
					      const double& b, 
					      const double& sigma){
    double A,B;
    double sigma_sqr = sigma*sigma;
    double aa = a*a;
    if (a==0.0){
	B = time;
	A = exp(sigma_sqr*pow(time,3))/6.0;
    }
    else {
	B = (1.0 - exp(-a*time))/a;
	A = exp( ((B-time)*(aa*b-0.5*sigma_sqr))/aa -((sigma_sqr*B*B)/(4*a)));
    };
    double dfact = A*exp(-B*r);
    return dfact;
}


double bond_option_price_call_zero_vasicek(const double& X,  // exercise price
					   const double& r, // current interest rate
					   const double& option_time_to_maturity,
					   const double& bond_time_to_maturity,
					   const double& a,  // parameters
					   const double& b, 
					   const double& sigma){  
    double T_t = option_time_to_maturity;
    double s_t = bond_time_to_maturity;
    double T_s = s_t-T_t;
    double v_t_T;
    double sigma_P;
    if (a==0.0) { 
	v_t_T = sigma * sqrt ( T_t ) ;
	sigma_P = sigma*T_s*sqrt(T_t); 
    }
    else {
	v_t_T = sqrt (sigma*sigma*(1-exp(-2*a*T_t))/(2*a));
	double B_T_s = (1-exp(-a*T_s))/a;
	sigma_P = v_t_T*B_T_s;
    };
    double h = (1.0/sigma_P) * log (term_structure_discount_factor_vasicek(s_t,r,a,b,sigma)/
	(term_structure_discount_factor_vasicek(T_t,r,a,b,sigma)*X) )
	+ sigma_P/2.0;
    double c = term_structure_discount_factor_vasicek(s_t,r,a,b,sigma)*N(h)
	-X*term_structure_discount_factor_vasicek(T_t,r,a,b,sigma)*N(h-sigma_P);
    return c;
};


