#include <iostream>
#include <cmath>              // mathematical C library
#include <algorithm>             // defining the max() operator
#include <vector>           // STL vector templates
using namespace std;
#include "finrecipes.h"


#ifndef PI 
const double PI=3.141592653589793238462643;
#endif 

double random_uniform_0_1(void){
    return double(rand())/double(RAND_MAX); // this uses the C library random number generator. 
};
double random_normal(void){
    double U1, U2, V1, V2;
    double S=2;
    while (S>=1) {
	U1 = random_uniform_0_1();
	U2 = random_uniform_0_1();
	V1 = 2.0*U1-1.0;
	V2 = 2.0*U2-1.0;
	S = pow(V1,2)+pow(V2,2);
    };
    double X1=V1*sqrt((-2.0*log(S))/S);
    return X1;
};



double simulate_lognormal_random_variable(const double& S,  // current value of variable
					  const double& r,  // interest rate
					  const double& sigma,  // volatitily 
					  const double& time) {  // time to final date
   double R = (r - 0.5 * pow(sigma,2) )*time;
   double SD = sigma * sqrt(time);
   return S * exp(R + SD * random_normal());
};

vector<double> 
simulate_lognormally_distributed_sequence(const double& S,  
					  const double& r,  
					  const double& sigma,
					  const double& time,  // time to final date
					  const int& no_steps){  // number of steps 
    vector<double> prices(no_steps);
    double delta_t = time/no_steps;
    double R = (r-0.5*pow(sigma,2))*delta_t;
    double SD = sigma * sqrt(delta_t);
    double S_t = S;                       // initialize at current price
    for (int i=0;  i<no_steps; ++i) {   
	S_t = S_t * exp(R + SD * random_normal());
	prices[i]=S_t;
    };
    return prices;
};

double option_price_put_european_simulated(const double& S,
					   const double& K,
					   const double& r,
					   const double& sigma,
					   const double& time,
					   const int& no_sims) {
    double sigma_sqr = sigma * sigma;
    double R = (r - 0.5 * sigma_sqr)*time;
    double SD = sigma * sqrt(time);
    double sum_payoffs = 0.0;
    for (int n=1; n<=no_sims; n++) {
	double S_T = S * exp(R + SD * random_normal());
	sum_payoffs += max(0.0, K-S_T);
    };
    return exp(-r*time) * (sum_payoffs/double(no_sims));
};

double option_price_delta_put_european_simulated(const double& S, 
						 const double& K, 
						 const double& r,
						 const double& sigma, 
						 const double& time,
						 const int& no_sims) {
    double sigma_sqr = sigma * sigma;
    double R = (r - 0.5 * sigma_sqr)*time;
    double SD = sigma * sqrt(time);
    double sum_payoffs = 0.0;
    double sum_payoffs_2 = 0.0;
    double q = S*0.01; 
    for (int n=1; n<=no_sims; n++) {
	double Z = random_normal();
	double S_T = S * exp(R + SD * Z);
	double S_T_2 = (S+q) * exp(R + SD * Z);
	sum_payoffs += max(0.0, K-S_T);
	sum_payoffs_2 += max(0.0, K-S_T_2);
    };
    double p = exp(-r*time) * ( sum_payoffs/no_sims); 
    double p2 = exp(-r*time) * ( sum_payoffs_2/no_sims); 
    return (p2-p)/q;
};
