#include <iostream>
using namespace std;
#include "finrecipes.h"



void test_option_price_call_black_scholes(){
    double S = 50; double K = 50; double r = 0.10;
    double sigma = 0.30; double time=0.50;
    cout << " Black Scholes call price = ";
    cout << option_price_call_black_scholes(S, K , r, sigma, time) << endl;
};
void test_black_scholes_partials_call(){
    cout  << " Black Scholes call partial derivatives " << endl;
    double S = 50; double K = 50; double r = 0.10;
    double sigma = 0.30; double time=0.50;
    double Delta, Gamma, Theta, Vega, Rho;
    option_price_partials_call_black_scholes(S,K,r,sigma, time,
					     Delta, Gamma, Theta, Vega, Rho);
    cout << "  Delta = " << Delta << endl;
    cout << "  Gamma = " << Gamma << endl;
    cout << "  Theta = " << Theta << endl;
    cout << "  Vega  = " << Vega << endl;
    cout << "  Rho   = " << Rho << endl;
};

void test_black_scholes_implied_volatility(){
    double S = 50; double K = 50; double r = 0.10; double time=0.50;
    double C=2.5;
    cout << " Black Scholes implied volatility using Newton search = ";
    cout << option_price_implied_volatility_call_black_scholes_newton(S,K,r,time,C) << endl;
};

void black_scholes_examples(){
    cout << "----------------------------------------" << endl;
    cout << "Examples in Black Scholes chapter " << endl; 
    cout << "----------------------------------------" << endl;
    test_option_price_call_black_scholes();
    test_black_scholes_partials_call();
    test_black_scholes_implied_volatility();
};
 
void test_analytical_geometric_average(){
    double S=100;  double K=100; double q=0;
    double r=0.06; double sigma=0.25; double time=1.0;
    cout << " Analytical geometric average = " 
	 << option_price_asian_geometric_average_price_call(S,K,r,q,sigma,time)
	 << endl;
};

void test_exotics_lookback(){
    double S=100; double Smin=S; double q = 0; double r = 0.06;
    double sigma = 0.346; double time = 1.0;
    cout << " Lookback call price = " 
	 << option_price_european_lookback_call(S,Smin,r,q,sigma,time) << endl;
};


void average_and_lookback_options_examples(){
    cout << "------------------------------------" << endl;
    cout << "Average and Lookback Options " << endl;
    cout << "------------------------------------" << endl;
    test_analytical_geometric_average();
    test_exotics_lookback();
 };

void test_heston(){
    double S=100,aux;
    double K=100;
    double r=0.01;
    double v=0.01;
    double tau=0.5;
    double rho=0;
    double kappa=2;
    double lambda=0.0;
    double theta=0.01;
    double sigma=0.01;
	//	for (int i=0;i<100;i++){
    cout << " Heston call price " << 
		heston_call_option_price( S, K, r, v, tau, rho, kappa, lambda, theta, sigma);
	//	/pow(1.2,i) << endl;
	//		S*=1.2;K*=1.2;
	//}
	};

void test_merton_jump_diff_call(){
    double S=100; double K=100; double r=0.05;
    double sigma=0.3; 
    double time_to_maturity=1;
    double lambda=0.5;
    double kappa=0.5;
    double delta=0.5;
    cout << " Merton Jump diffusion call = " 
	<< option_price_call_merton_jump_diffusion(S,K,r,sigma,time_to_maturity,lambda,kappa,delta)
	 << endl;
};

void alternative_formulas_examples(){
    cout << "-----------------------------" << endl;
    cout << "Alternative formulas " << endl;
    cout << "-----------------------------" << endl;
	test_heston();
	test_merton_jump_diff_call();
};

void test_explicit_finite_differences(){
    double S = 50.0;
    double K = 50.0;
    double r = 0.1; 
    double sigma = 0.4;
    double time=0.4167;
    int no_S_steps=20;
    int no_t_steps=11;
    cout << " explicit finite differences, european put price = ";
    cout << option_price_call_european_finite_diff_explicit(S,K,r,sigma,time,no_S_steps,no_t_steps) 
	 << endl;
    cout << " explicit finite differences, american put price = ";
    cout << option_price_put_american_finite_diff_explicit(S,K,r,sigma,time,no_S_steps,no_t_steps) 
	 << endl;
};

void finite_differences_examples(){
    cout << "----------------------------" << endl;
    cout << "Finite Differences examples " << endl;
    cout << "----------------------------" << endl;
    test_explicit_finite_differences();
};

void test_implicit_finite_differences(){
    double S = 50.0;
    double K = 50.0;
    double r = 0.1; 
    double sigma = 0.4;
    double time=0.5;
    int no_S_steps=200;
    int no_t_steps=200;
    cout << " black scholes put price = "  << option_price_put_black_scholes(S,K,r,sigma,time)<< endl;
    cout << " implicit American put price = ";
    cout << option_price_put_american_finite_diff_implicit(S,K,r,sigma,time,no_S_steps,no_t_steps) 
		<< endl;
};

void examples_finite_diffs_using_progonka(){
    cout << "----------------------------" << endl;
    cout << "Implicit Finite Differences examples" << endl;
    cout << "----------------------------" << endl;
    test_implicit_finite_differences();
};

void test_binomial_approximations_option_pricing(){
    double S = 100.0;    double K = 100.0;
    double r = 0.1;      double sigma = 0.25;
    double time=1.0;
    int no_steps = 100;
    cout << " european call = " 
	 << option_price_call_european_binomial(S,K,r,sigma,time,no_steps) 
	 << endl;
};

 
void binomial_approximations_examples(){
    cout << "-------------------------------------" << endl;
    cout << "Binomial Approximations examples" << endl;
    cout << "-------------------------------------" << endl;
    test_binomial_approximations_option_pricing();
};
 
void test_trinomial_trees(){
    double S = 100.0; double K = 100.0;
    double r = 0.1;   double sigma = 0.25;
    double time=1.0;  double q=0;  
    int no_steps = 100;
    cout << " american put using trinomial = " 
	 << option_price_put_american_trinomial(S,K,r,q,sigma,time,no_steps) 
	 << endl;
}; 

void examples_trinomial_trees(){
    cout << "----------------------------" << endl;
    cout << "-- trinomial trees ---------" << endl;
    cout << "----------------------------" << endl;
    test_trinomial_trees();
};





void test_cumulative_normal() {
    cout << " N(0) = " << N(0) << endl;
    cout << " N(0,0,0) = " << N(0,0,0) << endl;
};

void test_random_normal(){
    cout << " 5 random uniform numbers between 0 and 1: " << endl << "    "; 
    for (int i=0;i<5;++i){ cout << " " << random_uniform_0_1(); }; cout << endl;
    cout << " 5 random normal(0,1) numbers: " << endl << "    ";
    for (int i=0;i<5;++i){ cout << " " << random_normal() ; }; cout << endl;
};

void normal_distribution_examples(){
    cout << "----------------------------------" << endl;
    cout << " Normal distribution calculations " << endl;
    cout << "----------------------------------" << endl;
    test_cumulative_normal();
    test_random_normal();
};

void test_simulation_pricing() { 
    double S=100.0;  double K=100.0; double r=0.1; double sigma=0.25;
    double time=1.0; int no_sims=5000;
    cout << " put:  black scholes price = " << option_price_put_black_scholes(S,K,r,sigma,time) << endl;
    cout << "       simulated price     = " 
	 << option_price_put_european_simulated(S,K,r,sigma,time, no_sims)  << endl;
};

void test_simulation_pricing_delta(){
    double S=100.0;  double K=100.0; double r=0.1; double sigma=0.25;
    double time=1.0; int no_sims=5000;
    cout << " put: bs delta  = " <<  option_price_delta_put_black_scholes(S,K,r,sigma,time) 
	 << "      sim delta = " <<  option_price_delta_put_european_simulated(S,K,r,sigma,time,no_sims) 
	 << endl;
};
 


void test_simulation_bs_case_using_generic_routine(){
    double S  = 100; double X = 100; double r = 0.1;
    double sigma = 0.25; double time  = 1.0;  int no_sims     = 50000;
    cout << "Black Scholes call option price = " << option_price_call_black_scholes(S,X,r,sigma,time) << endl;
    cout << "Simulated call option price     = " 
	 << derivative_price_simulate_european_option_generic(S,X,r,sigma,time,payoff_call,no_sims)
	 << endl;
    cout << "Black Scholes put option price  = " << option_price_put_black_scholes(S,X,r,sigma,time) << endl;
    cout << "Simulated put option price      = " 
	 << derivative_price_simulate_european_option_generic(S,X,r,sigma,time,payoff_put,no_sims)
	 << endl;
};

void test_simulation_bs_case_using_generic_routine_improving_efficiency(){
    double S = 100; double K = 100; double r = 0.1;
    double sigma    = 0.25; double time = 1; int no_sims     = 50000;
    cout << "Black Scholes call option price = " 
	 << option_price_call_black_scholes(S,K,r,sigma,time) << endl;
    cout << "Simulated call option price     = " 
	 << derivative_price_simulate_european_option_generic(S,K,r,sigma,time, payoff_call,no_sims)
	 << endl;
    cout << "Simulated call option price, CV = " 
	 << derivative_price_simulate_european_option_generic_with_control_variate(S,K,r,sigma,time,
										   payoff_call,no_sims)
	 << endl;
    cout << "Simulated call option price, AV = " 
	 << derivative_price_simulate_european_option_generic_with_antithetic_variate(S,K,r,sigma,time,
										      payoff_call,no_sims)
	 << endl;
    cout << "Black Scholes put option price  = " << option_price_put_black_scholes(S,K,r,sigma,time) << endl;
    cout << "Simulated put option price      = " 
	 << derivative_price_simulate_european_option_generic(S,K,r,sigma,time,payoff_put,no_sims) << endl;
    cout << "Simulated put option price,  CV = " 
	 << derivative_price_simulate_european_option_generic_with_control_variate(S,K,r,sigma,time,
										   payoff_put,no_sims)
	 << endl;
    cout << "Simulated put option price,  AV = " 
	 << derivative_price_simulate_european_option_generic_with_antithetic_variate(S,K,r,sigma,time,
										      payoff_put,no_sims)
	 << endl;
};

void test_simulation_binary_options(){
    double S=100.0; double K=100.0; double r=0.1; double sigma=0.25;
    double time=1.0;  int no_sims=5000;
    cout << " cash or nothing, Q=1: " 
	 << derivative_price_simulate_european_option_generic(S,K,r,sigma,time,
							      payoff_cash_or_nothing_call,
							      no_sims)
	 << endl;
    cout << " control_variate " 
	 << derivative_price_simulate_european_option_generic_with_control_variate(S,K,r,sigma,time,
										   payoff_cash_or_nothing_call,
										   no_sims)
	 << endl;
    cout << " antithetic_variate " 
	 << derivative_price_simulate_european_option_generic_with_antithetic_variate(S,K,r,sigma,time,
										      payoff_cash_or_nothing_call,
										      no_sims)
	 << endl;
    cout << " asset or nothing: " 
	 << derivative_price_simulate_european_option_generic(S,K,r,sigma,time,
							      payoff_asset_or_nothing_call,
							      no_sims)
	 << endl;
    cout << " control_variate " 
	 << derivative_price_simulate_european_option_generic_with_control_variate(S,K,r,sigma,time,
										   payoff_asset_or_nothing_call,
										   no_sims)
	 << endl;
    cout << " antithetic_variate " 
	 << derivative_price_simulate_european_option_generic_with_antithetic_variate(S,K,r,sigma,time,
										      payoff_asset_or_nothing_call,
										      no_sims)
	 << endl;
};


void test_simulate_general_european(){
    cout << "Testing general simulation of European options " << endl;
    double S=100; double K=120; double r = 0.10;
    double time = 1.0; double sigma = 0.25; int no_sims = 10000; int no_steps = 250;
    double q=0;

    cout << " simulated arithmetric average " 
	 << " S= " <<  S << " r= " << r << " price="
	 << derivative_price_simulate_european_option_generic(S,K,r,sigma,time,
							      payoff_arithmetric_average_call,
							      no_steps,no_sims) 
	 << endl; 

    cout << " simulated geometric average = " 
	 << derivative_price_simulate_european_option_generic(S,K,r,sigma,time,
							      payoff_geometric_average_call,
							      no_steps,no_sims) 
	 << endl; 
    cout << " analytical lookback call = " 
	 << option_price_european_lookback_call(S,S,r,q,sigma,time)
	 << endl;
    cout << " simulated lookback put = " 
	 << derivative_price_simulate_european_option_generic(S,0,r,sigma,time,
							      payoff_lookback_put,
							      no_steps,no_sims) 
	 << endl; 
    cout << " analytical lookback call = " 
	 << option_price_european_lookback_call(S,S,r,q,sigma,time)
	 << endl;

    cout << " simulated lookback call = " 
	 << derivative_price_simulate_european_option_generic(S,0,r,sigma,time,
							      payoff_lookback_call,
							      no_steps,no_sims) 
	 << endl; 
    cout << " simulated lookback call using control variates = " 
	 << derivative_price_simulate_european_option_generic_with_control_variate(S,0,r,sigma,time,
										   payoff_lookback_call,
										   no_steps,no_sims) 
	 << endl; 
};

void simulation_examples(){
    cout << "--------------------------" << endl;
    cout << " Simulation examples " << endl;
    cout << "--------------------------" << endl;
    test_simulation_pricing();
    test_simulation_pricing_delta(); 
	test_simulation_bs_case_using_generic_routine();
    test_simulation_bs_case_using_generic_routine_improving_efficiency();
    test_simulation_binary_options();
	test_simulate_general_european();

};

int main() {
    black_scholes_examples();//dt Grecs et vol impl
	average_and_lookback_options_examples();
	alternative_formulas_examples();//heston et merton
	finite_differences_examples();//expl
	examples_finite_diffs_using_progonka();
    binomial_approximations_examples();
	examples_trinomial_trees();
	simulation_examples();//prix, delta, redvar, also binaire..
    normal_distribution_examples();//cumulative and random
};
